"""
Render CLDF Markdown to common markdown.

The CLDF Markdown text can be read
- from a parameter, via `--text-string`
- from a file, specified as `--text-file`
- from a media item associated with a CLDF dataset via `--media-id`.
"""
import re
import pathlib
import argparse

from clldutils.clilib import PathType, ParserError
from clldutils.markup import MarkdownImageLink
from pycldf.ext import discovery
from pycldf.ext.markdown import DatasetMapping
from pycldf.media import MediaTable
from termcolor import colored

from cldfviz.text import iter_templates, render, iter_cldfviz_links
from cldfviz.cli_util import add_testable
from . import map, tree


def get_dataset(locator):
    prefix = None
    if ':' in locator and not locator.startswith('http'):
        prefix, _, locator = locator.partition(':')
        if not DatasetMapping.key_pattern.fullmatch(prefix):
            raise ParserError('Invalid dataset prefix: {}'.format(prefix))  # pragma: no cover
    return prefix, locator


def register(parser):
    add_testable(parser)
    parser.add_argument('datasets', type=get_dataset, nargs='+')
    parser.add_argument('-l', '--list', help='list templates', default=False, action='store_true')
    parser.add_argument(
        '--media-id', default=None)
    parser.add_argument(
        '--text-string', default=None)
    parser.add_argument(
        '--text-file', type=PathType(type='file', must_exist=True), default=None)
    parser.add_argument('--templates', type=PathType(type='dir'), default=None)
    parser.add_argument('--output', type=PathType(type='file', must_exist=False), default=None)
    parser.add_argument('--download-dir', type=PathType(type='dir'), default=None)


def run(args):
    dss = {
        prefix: discovery.get_dataset(locator, args.download_dir)
        for prefix, locator in args.datasets}

    if args.list:
        print(colored('Available templates:', attrs=['bold', 'underline']) + '\n')
        for p, doc, vars in iter_templates():
            component, _, type_ = p.stem.partition('_')
            if (component == 'Source' and any(ds.sources for ds in dss.values())) or \
                    any(component in ds for ds in dss.values()):
                print(colored('{} {}'.format(component, type_), attrs=['bold']))
                print('Usage: ' + colored('[<label>]({}{}#cldf:{})'.format(
                    component,
                    '?var1&var2' if vars else '',
                    '__all__' if type_ == 'index' else '<object-ID>'), color='blue'))
                if vars:
                    print('Variables: ' + colored(', '.join(vars), color='blue'))
                if doc:
                    print(doc)
        return

    assert args.text_string or args.text_file or args.media_id
    text = args.text_string
    if (not text) and args.text_file:
        text = args.text_file.read_text(encoding='utf8')
    if not text:
        prefix, _, mid = args.media_id.partition(':')
        if not mid:
            mid, prefix = prefix, None
        for media in MediaTable(dss[prefix]):
            if media.id == mid:
                assert media.mimetype == 'text/markdown'
                # FIXME: Also check for conformsTo == CLDF Markdown column?
                text = media.read()
                break

    res = render(text, dss, args.templates)
    if args.output:
        args.output.parent.mkdir(parents=True, exist_ok=True)
        args.output.write_text(res, encoding='utf8')
        args.log.info('{} written'.format(args.output))

    create_images(
        args,
        res,
        dss,
        args.output.parent if args.output else
        (pathlib.Path('.') if args.text_string or (not args.text_file) else args.text_file.parent))

    def clean(ml):
        if re.match(r'cldfviz\.(map|tree)', ml.parsed_url.fragment):
            ml.update_url(query='')
        return ml

    res = MarkdownImageLink.replace(res, clean)

    if not args.output:
        print(res)


def create_images(oargs, md, dss, base_dir):
    for prefix, ds in dss.items():
        for ml in iter_cldfviz_links(md):
            if prefix is None or (ml.parsed_url.fragment.partition('-')[2] == prefix):
                p = base_dir.joinpath(ml.parsed_url.path)
                p.parent.mkdir(parents=True, exist_ok=True)
                args = [str(ds.tablegroup._fname)]
                kw = ml.parsed_url_query
                kw['output'] = [str(p)]
                for k, v in kw.items():
                    if v[0]:
                        args.extend(['--' + k, v[0]])
                    else:
                        args.append('--' + k)

                if 'tree' in ml.parsed_url.fragment:
                    cmd = tree
                elif 'map' in ml.parsed_url.fragment:
                    kw['format'] = [p.suffix[1:].lower()]
                    cmd = map
                else:
                    raise NotImplementedError(ml.parsed_url.fragment)  # pragma: no cover

                p = argparse.ArgumentParser()
                cmd.register(p)
                args = p.parse_args(args)
                if oargs.test:
                    args.test = True
                args.log = oargs.log
                cmd.run(args)



import re
import json
import inspect
import pathlib
import argparse
import webbrowser

from clldutils.text import split_text_with_context
from clldutils.clilib import PathType
from clldutils import path


def add_jinja_template(parser, default):
    parser.add_argument(
        "--template",
        type=PathType(type='file'),
        default=default,
        help="Template file using Jinja2 syntax (see https://jinja.palletsprojects.com/). "
             "To create a custom template, you might want to start out with a copy of the "
             "default template.",
    )


def add_open(parser):
    parser.add_argument(
        "--open", action='store_true', default=False,
        help="Open the output file in the browser. (Requires specifying --output as well.)"
    )
    try:
        parser.add_argument(
            "-o", "--output", type=PathType(type="file", must_exist=False), default=False,
            help="(non-existing) path name to write the result to.",
        )
    except argparse.ArgumentError:  # pragma: no cover
        pass  # output option already added.


def open_output(args):
    if args.output and args.open and not getattr(args, 'test', False):  # pragma: no cover
        webbrowser.open(args.output.resolve().as_uri(), new=1)


def write_output(args, res):
    if args.output:
        args.output.write_text(res, encoding="utf8")
        print("Output written to {}".format(args.output))
        open_output(args)
    else:
        print(res)


def add_language_filter(parser):
    parser.add_argument(
        '--language-filters',
        default=None,
        help="JSON object specifying filter criteria for included languages. Keys must be "
             "names of columns in the datasets' LanguageTable, values are interpreted as "
             "regular expressions if they are strings or as literal values otherwise and will "
             "be matched against the value of a language for the specified column. Only "
             "languages matching all criteria are included in the analysis.",
    )


def get_language_filter(args):
    if args.language_filters is None:
        return

    def language_filter(lg):
        for k, v in json.loads(args.language_filters).items():
            val = lg.data[k]
            if isinstance(v, str):
                if isinstance(val, list):
                    if v not in val:
                        return False
                elif not re.search(v, val or ''):
                    return False
            else:
                if lg.data[k] != v:
                    return False
        return True
    return language_filter


def add_testable(parser):
    parser.add_argument('--test', action='store_true', default=False, help=argparse.SUPPRESS)


def add_listvalued(parser, *args, **kw):
    kw.setdefault('default', [])
    kw.setdefault('type', lambda s: split_text_with_context(s, ',', brackets={'{': '}'}))
    parser.add_argument(*args, **kw)


def import_module(dotted_name_or_path):
    import importlib
    p = pathlib.Path(dotted_name_or_path)
    if p.exists():
        return path.import_module(p.resolve())
    return importlib.import_module(dotted_name_or_path)


def import_subclass(dotted_name_or_path, cls):
    mod = import_module(dotted_name_or_path)
    for _, obj in inspect.getmembers(mod):
        if inspect.isclass(obj) and issubclass(obj, cls) and not obj.__subclasses__():
            return obj



"""
We provide a Glottolog object that abstracts whether the data is accessed via pyglottolog.Glottolog
from glottolog/glottolog-like data or via pycldf.Dataset from glottolog/glottolog-cldf-like
data.
"""
import argparse
import collections

import attr
from pycldf import Dataset
from pycldf.ext import discovery
from cldfbench.cli_util import add_catalog_spec, IGNORE_MISSING
from clldutils.clilib import PathType
import newick

try:
    import pyglottolog
except ImportError:  # pragma: no cover
    pyglottolog = None

__all__ = ['Glottolog', 'Languoid']


@attr.s
class Languoid:
    id = attr.ib()
    name = attr.ib()
    lat = attr.ib()
    lon = attr.ib()

    @classmethod
    def from_dict(cls, d):
        return cls(id=d['id'], name=d['name'], lat=d['latitude'], lon=d['longitude'])

    @classmethod
    def from_languoid(cls, lang):
        return cls(id=lang.id, name=lang.name, lat=lang.latitude, lon=lang.longitude)


class Glottolog(collections.UserDict):
    def __init__(self, api_or_dataset):
        self.api = api_or_dataset
        super().__init__()
        if isinstance(self.api, Dataset):
            for row in self.api.iter_rows('LanguageTable', 'id', 'name', 'latitude', 'longitude'):
                self[row['id']] = Languoid.from_dict(row)
        else:
            for lang in self.api.languoids():
                self[lang.id] = Languoid.from_languoid(lang)

    @staticmethod
    def add(parser):
        add_catalog_spec(parser, 'glottolog', default=IGNORE_MISSING)
        parser.add_argument(
            '--glottolog-cldf',
            default=None,
            help="Dataset locator for the glottolog-cldf dataset.")
        try:
            parser.add_argument(
                '--download-dir',
                type=PathType(type='dir'),
                help='An existing directory to use for downloading a dataset (if necessary).',
                default=None,
            )
        except argparse.ArgumentError:
            pass

    @classmethod
    def from_args(cls, args):
        if args.glottolog_cldf:
            return cls(discovery.get_dataset(args.glottolog_cldf, download_dir=args.download_dir))
        if args.glottolog:
            if hasattr(args.glottolog, 'api'):
                # cldfbench has already initialized a pyglottolog.Glottolog instance!
                return cls(args.glottolog.api)
            if args.glottolog != IGNORE_MISSING:
                assert pyglottolog
                return cls(pyglottolog.Glottolog(args.glottolog))

    def newick(self, gc):
        if isinstance(self.api, Dataset):
            for row in self.api.iter_rows(
                    'ValueTable', 'languageReference', 'parameterReference', 'value'):
                if row['languageReference'] == gc and \
                        row['parameterReference'] == 'subclassification':
                    return newick.loads(row['value'])[0]
        else:
            return self.api.languoid(gc).newick_node(template="{l.id}")



"""
Plots a phylogeny as SVG.
"""
import pathlib

from pycldf.cli_util import add_dataset, get_dataset
from pycldf.trees import TreeTable
from cldfbench.cli_util import add_catalog_spec, IGNORE_MISSING

from cldfviz.cli_util import (
    add_testable, add_language_filter, get_language_filter, add_open, write_output,
)
from cldfviz.tree import render


def register(parser):
    add_testable(parser)
    add_dataset(parser)
    add_language_filter(parser)
    add_catalog_spec(parser, 'glottolog', default=IGNORE_MISSING)
    parser.add_argument(
        '--glottolog-links',
        help="Turn language labels into links to Glottolog (where possible).",
        action='store_true',
        default=False)
    parser.add_argument(
        '--ascii-art',
        help="Only print tree as ASCII graphic to the terminal.",
        action='store_true',
        default=False)
    parser.add_argument(
        '--name-as-label',
        action='store_true',
        default=False)
    parser.add_argument('--title', default=None)
    parser.add_argument('--width', type=int, default=500)
    parser.add_argument(
        '--tree-id',
        help="ID in TreeTable of the dataset, identifying the tree to plot.",
        default=None)
    parser.add_argument(
        '--styles',
        help="Python dict suitable to pass into `toytree.tree.draw` (or path to a text file "
             "containing such a dict). "
             "See https://toytree.readthedocs.io/en/latest/8-styling.html#",
        default='{}')
    add_open(parser)


def run(args):
    cldf = get_dataset(args)
    res = None

    for tree in TreeTable(cldf):
        if args.tree_id is None or (args.tree_id == tree.id):
            if args.ascii_art:
                print(tree.newick(strip_comments=True).ascii_art())
                return
            if args.title:
                legend = args.title
            else:
                legend = tree.name
                dcol = cldf.get(('TreeTable', 'description'))
                if dcol:
                    legend += ' - {}'.format(tree.row[dcol.name])

                if tree.tree_branch_length_unit:
                    legend += ' with branches in {}'.format(tree.tree_branch_length_unit)
            glangs = {}
            if args.glottolog and args.glottolog_links:  # pragma: no cover
                glangs = {lg.id: lg.name for lg in args.glottolog.api.languoids()}
            if pathlib.Path(args.styles).exists():
                args.styles = pathlib.Path(args.styles).read_text(encoding='utf8')
            lf = get_language_filter(args)
            res = render(
                tree,
                glottolog_mapping={
                    r['id']: (r['glottocode'], glangs.get(r['glottocode']) or '') for r in
                    cldf.iter_rows('LanguageTable', 'id', 'glottocode') if r['glottocode']},
                legend=legend,
                width=args.width,
                styles=eval(args.styles),
                with_glottolog_links=args.glottolog_links,
                labels={
                    r['id']: r['name']
                    for r in cldf.iter_rows('LanguageTable', 'id', 'name')
                } if args.name_as_label else None,
                leafs=[lg.id for lg in cldf.objects('LanguageTable') if lf(lg)] if lf else None,
            )
            break
    else:
        raise ValueError('no matching tree found!')  # pragma: no cover
    write_output(args, res)



import json
import string

import attr
from clldutils import svg
from clldutils.html import HTML

from .base import Map, PACIFIC_CENTERED
import cldfviz

SHAPE_MAP = {
    'triangle_down': 'f',
    'triangle_up': 't',
    'square': 's',
    'diamond': 'd',
    'circle': 'c',
}
BASE_LAYERS = {
    'OpenStreetMap': (
        'https://tile.openstreetmap.org/{z}/{x}/{y}.png',
        {
            'maxZoom': 18,
            'attribution':
                '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>'}
    ),
    'OpenTopoMap': (
        'https://{s}.tile.opentopomap.org/{z}/{x}/{y}.png',
        {
            'maxZoom': 17,
            'attribution':
                'Map data: &copy; <a href="https://www.openstreetmap.org/copyright">OpenStreetMap'
                '</a> contributors, <a href="http://viewfinderpanoramas.org">SRTM</a> | '
                'Map style: &copy; <a href="https://opentopomap.org">OpenTopoMap</a> '
                '(<a href="https://creativecommons.org/licenses/by-sa/3.0/">CC-BY-SA</a>)'}
    ),
    'USGS.USTopo': (
        'https://basemap.nationalmap.gov/arcgis/rest/services/USGSTopo/MapServer/tile/{z}/{y}/{x}',
        {
            'maxZoom': 20,
            'attribution':
                'Tiles courtesy of the <a href="https://usgs.gov/">U.S. Geological Survey</a>'
        }
    ),
    'USGS.USImagery': (
        'https://basemap.nationalmap.gov/arcgis/rest/services/'
        'USGSImageryOnly/MapServer/tile/{z}/{y}/{x}',
        {
            'maxZoom': 20,
            'attribution':
                'Tiles courtesy of the <a href="https://usgs.gov/">U.S. Geological Survey</a>'
        }
    ),
    'USGS.USImageryTopo': (
        'https://basemap.nationalmap.gov/arcgis/rest/services/'
        'USGSImageryTopo/MapServer/tile/{z}/{y}/{x}',
        {
            'maxZoom': 20,
            'attribution':
                'Tiles courtesy of the <a href="https://usgs.gov/">U.S. Geological Survey</a>'
        }
    ),
}


@attr.s
class LeafletMarkerSpec:
    icon = attr.ib(default=svg.data_url(svg.icon('c000')))
    name = attr.ib(default=None)
    values = attr.ib(default=None)
    tooltip = attr.ib(default=None)
    tooltip_class = attr.ib(default=None)
    markersize = attr.ib(default=None)
    css = attr.ib(default=None)


class MapLeaflet(Map):
    __formats__ = ['html']
    __marker_class__ = LeafletMarkerSpec

    def __init__(self, languages, args):
        Map.__init__(self, languages, args)
        self.features = []
        self.legend = ''
        self.css = set()

    @staticmethod
    def add_options(parser, help_suffix):
        parser.add_argument(
            '--base-layer',
            default='OpenStreetMap',
            help="Tile layer for Leaflet maps. {}".format(help_suffix),
            choices=list(BASE_LAYERS),
        )
        parser.add_argument(
            '--with-layers',
            default=False,
            action='store_true',
            help='Create clickable Leaflet layers for each parameter value. {}'.format(help_suffix),
        )
        parser.add_argument(
            '--with-layers-for-combinations',
            default=False,
            action='store_true',
            help='Create clickable Leaflet layers for each combination of parameter values. '
                 '{}'.format(help_suffix),
        )
        parser.add_argument(
            '--value-template',
            default='{parameter}: {code}',
            help='A format string specifying how to format values as text labels.',
        )

    def _lonlat(self, language):
        lon, lat = language.lon, language.lat
        if self.args.pacific_centered and lon <= PACIFIC_CENTERED - 180:
            # Anything west of 26°W is moved by 360°.
            lon += 360  # make the map pacific-centered.
        return [lon, lat]

    def _icon(self, colors):
        res = self.get_shape_and_color(colors)
        if res:
            return svg.icon(SHAPE_MAP[res[0]] + res[1].replace('#', ''))
        return svg.pie([1] * len(colors), colors, stroke_circle=True)

    def add_language(self, language, values, colormaps, spec=None):
        icon = self._icon([colormaps[pid](vals[0].v) for pid, vals in values.items()])
        props = {
            "name": language.name,
            "tooltip": language.name,
            "values": ' / '.join(
                [self.args.value_template.format(
                    parameter=pid, code=vals[0].v) for pid, vals in values.items() if vals]),
            "icon": svg.data_url(icon),
            "markersize": self.args.markersize,
            "tooltip_class": "tt",
        }
        if spec:
            if spec.css:
                self.css.add(spec.css)
            props.update(
                {k: v for k, v in attr.asdict(spec).items() if v is not None and k != 'css'})
        self.features.append({
            # A language as GeoJSON point with svg marker icon
            "geometry": {"coordinates": self._lonlat(language), "type": "Point"},
            "id": language.id,
            "properties": props,
            "type": "Feature"
        })

    def add_legend(self, parameters, colormaps):
        def marker(colors):
            return HTML.img(
                src=svg.data_url(self._icon(colors)),
                width="{}".format(min([20, self.args.markersize * 2])))

        trs = []
        for i, (pid, parameter) in enumerate(parameters.items()):
            if i != 0:
                trs.append(HTML.tr(HTML.th(HTML.hr(), colspan='2')))
            trs.append(HTML.tr(
                HTML.th(
                    marker(['#000000' if j == i else '#ffffff' for j in range(len(parameters))])),
                HTML.th(parameter.name, style="text-align: left;")
            ))
            if isinstance(parameter.domain, tuple):
                # Create an HTML color bar for a continuous variable, as table with two rows and
                # 11 columns.
                min_, max_ = parameter.domain
                tds_label = []
                tds_color = []
                for j in range(11):
                    if j == 0:
                        tds_label.append(HTML.td(str(round(min_, 2))))
                    elif j == 10:
                        tds_label.append(HTML.td(str(round(max_, 2)), style="text-align: right;"))
                    else:
                        tds_label.append(HTML.td(' '))
                    tds_color.append(HTML.td(
                        ' ',
                        style='height: 20px; width: 1em; background-color: {};'.format(
                            colormaps[pid](min_ + j * (max_ - min_) / 10))))
                trs.append(HTML.tr(HTML.td(HTML.table(
                    HTML.tr(*tds_label),
                    HTML.tr(*tds_color)
                ), colspan='2')))
            else:
                for v, label in parameter.domain.items():
                    trs.append(HTML.tr(
                        HTML.td(marker([
                            colormaps[pid](v) if j == i else '#ffffff'
                            for j in range(len(parameters))])),
                        HTML.td(str(label))
                    ))
        self.legend = HTML.table(*trs, **{'class': 'legend'})

    def __exit__(self, exc_type, exc_val, exc_tb):
        """write files"""
        html = string.Template(
            cldfviz.PKG_DIR.joinpath('templates', 'map', 'leaflet.html').read_text(encoding='utf8')
        ).substitute(
            title=self.args.title or '',
            css='\n'.join(sorted(self.css)),
            legend=self.legend,
            options=json.dumps({
                'language_labels': self.args.language_labels,
                'with_layers': self.args.with_layers,
                'with_layers_for_combinations': self.args.with_layers_for_combinations,
            }),
            tile_url=json.dumps(BASE_LAYERS[self.args.base_layer][0]),
            tile_options=json.dumps(BASE_LAYERS[self.args.base_layer][1]),
            geojson=json.dumps({"features": self.features, "type": "FeatureCollection"}))
        self.args.output.write_text(html, encoding='utf8')



"""
Build an HTML file to display a wordlist for a particular concept with associated audio.

Assumes that
- the dataset is a CLDF Wordlist, coding concepts as parameters,
- forms are linked to media files
  - via a column with propertyUrl "mediaReference" in FormTable or
  - a column with propertyUrl "formReference" in MediaTable.
"""
import collections

from clldutils.clilib import PathType
from pycldf.terms import term_uri
from pycldf.cli_util import get_dataset, add_dataset
from pycldf.media import MediaTable

from cldfviz.cli_util import add_open, write_output, add_jinja_template
from cldfviz.template import render_jinja_template, TEMPLATE_DIR


def as_list(obj):
    if isinstance(obj, list):
        return obj
    return [obj]


def register(parser):
    add_dataset(parser)
    parser.add_argument(
        'concept',
        help='Specify a concept in the form COLNAME=VALUE or cldf:PROPNAME=VALUE, e.g. '
             '"Name=hand" or "cldf:id=10", using column names used in ParameterTable or '
             'CLDF properties available in ParameterTable.')
    # parser.add_argument('--mimetype', default=None)
    mod = __name__.split('.')[-1]
    add_jinja_template(parser, TEMPLATE_DIR / mod / '{}.html'.format(mod))
    parser.add_argument(
        '--media-dir',
        help="If media files are available locally (downloaded via pycldf's `cldf downloadmedia` "
             "command), the media directory can be passed to access the audio files on disk.",
        type=PathType(type='dir'),
        default=None)
    add_open(parser)


def run(args):
    ds = get_dataset(args)

    # Determine the relevant concept (aka parameter):
    colspec, sep, match = args.concept.partition('=')
    colname = None
    if not sep:
        match = colspec
    else:
        cldfprop = colspec.startswith('cldf:')
        colspec = colspec.replace('cldf:', '')
        colname = ds['ParameterTable', term_uri(colspec)].name if cldfprop else colspec

    if 'ParameterTable' in ds:
        for row in ds.iter_rows('ParameterTable', 'id', 'name'):
            if (colname and row[colname] == match) or row['id'] == match:
                pid = row['id']
                break
        else:
            raise ValueError(args.concept)  # pragma: no cover
    else:
        assert colname is None
        pid = match

    # Get relevant forms and linked media:
    forms, media = [], set()
    if ('FormTable', 'mediaReference') in ds:
        for form in ds.objects('FormTable'):
            if form.cldf.parameterReference == pid:
                mrefs = as_list(form.cldf.mediaReference)
                media |= set(mrefs)
                forms.append((form, mrefs))
    elif ('MediaTable', 'formReference') in ds:
        media_by_fid = collections.defaultdict(list)
        for row in ds.iter_rows('MediaTable', 'id', 'formReference'):
            for fid in as_list(row['formReference']):
                media_by_fid[fid].append(row['id'])
        for form in ds.objects('FormTable'):
            if form.cldf.parameterReference == pid:
                mrefs = media_by_fid.get(form.id, [])
                media |= set(mrefs)
                forms.append((form, mrefs))

    media = {mid: None for mid in media}

    # Retrieve relevant media - filtered by media type:
    for file in MediaTable(ds):
        if file.id in media and file.mimetype.type == 'audio':  # filter!
            if args.media_dir:
                if file.local_path(args.media_dir).exists():
                    # Read audio from the file system:
                    media[file.id] = 'file://{}'.format(
                        file.local_path(args.media_dir).resolve())
            else:
                # Read audio from the URL:
                media[file.id] = file.url

    res = render_jinja_template(
        args.template,
        ds=ds,
        pid=pid,
        parameter=forms[0][0].parameter,
        forms=[(form, [media[mid] for mid in mrefs if media[mid]]) for form, mrefs in forms],
        local=bool(args.media_dir))
    write_output(args, res)



import re

from pycldf import Dataset
from pycldf.ext.markdown import CLDFMarkdownText
import jinja2
import jinja2.meta
from clldutils.misc import nfilter
from clldutils.markup import MarkdownLink, MarkdownImageLink

import cldfviz

__all__ = ['iter_templates', 'render', 'iter_cldfviz_links']

TEMPLATE_DIR = cldfviz.PKG_DIR.joinpath('templates', 'text')


def _add_filters(env):
    def paragraphs(s):
        return '\n\n'.join(s.split('\n'))

    env.filters['paragraphs'] = paragraphs


def get_env(template_dir=None, fallback_template_dir=None):
    loader = jinja2.FileSystemLoader(
        searchpath=[str(d) for d in nfilter([template_dir, fallback_template_dir, TEMPLATE_DIR])])
    env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
    _add_filters(env)
    return env


def iter_templates():
    env = get_env()
    for p in sorted(TEMPLATE_DIR.iterdir(), key=lambda pp: pp.name):
        m = re.match(r"{#(.+?)#}", p.read_text(encoding='utf8'), flags=re.MULTILINE | re.DOTALL)
        doc = m.group(1) if m else None
        parsed_content = env.parse(env.loader.get_source(env, p.name))
        vars = jinja2.meta.find_undeclared_variables(parsed_content)
        yield p, doc, [v for v in vars if v != 'ctx']


def pad_ex(obj, gloss):
    out_obj = []
    out_gloss = []
    for o, g in zip(obj, gloss):
        g = g or ""
        diff = len(o) - len(g)
        if diff < 0:
            o += " "*-diff  # noqa E225
        else:
            g += " " * diff
        out_obj.append(o)
        out_gloss.append(g)
    return "  ".join(out_obj).strip(), "  ".join(out_gloss).strip()


def render(doc, cldf_dict, template_dir=None, loader=None, func_dict=None):
    """
    Render CLDF Markdown using customizable jinja2 templates.

    Features:

    - Reference list: Include a list of cited references using the link \
      `[](Source?cited_only#cldf:__all__)`

    :param doc:
    :param cldf_dict:
    :param template_dir:
    :param loader:
    :param func_dict:
    :return:
    """
    func_dict = func_dict or {}
    func_dict.update({"pad_ex": pad_ex})

    if isinstance(cldf_dict, Dataset):
        cldf_dict = {None: cldf_dict}

    folder_env = get_env(template_dir=template_dir)
    if loader is None:
        env = folder_env
    else:
        env = jinja2.Environment(
            loader=jinja2.ChoiceLoader([loader, folder_env.loader]),
            trim_blocks=True,
            lstrip_blocks=True)
        _add_filters(env)

    proc = TemplateRenderer(
        env, func_dict, doc, cldf_dict,
        # download_dir!
    )
    return proc.render()


def iter_cldfviz_links(md):
    for match in MarkdownImageLink.pattern.finditer(md):
        ml = MarkdownImageLink.from_match(match)
        if re.match(r'cldfviz\.(map|tree)', ml.parsed_url.fragment):
            yield ml


class TemplateRenderer(CLDFMarkdownText):
    def __init__(self, env, func_dict, *args, **kw):
        super().__init__(*args, **kw)
        self.env = env
        self.func_dict = func_dict
        self.with_partial_local_reflist = False
        self.cited = None

    def render_template(self, fname_or_component, ctx, index=False, fmt='md'):
        # Determine the template to use ...
        tmpl_fname = ctx.pop(
            '__template__',  # ... by looking for an explicit name ...
            # ... and falling back to the "most suitable" one.
            '{}_{}.{}'.format(fname_or_component, 'index' if index else 'detail', fmt),
        )
        jinja_template = self.env.get_template(tmpl_fname)
        jinja_template.globals.update(self.func_dict)
        return jinja_template.render(**ctx)

    def get_tmpl_context(self, ml):
        tmpl_context = {k: True if v[0] == '' else v[0] for k, v in ml.parsed_url_query.items()}
        for k in tmpl_context:
            if k.startswith('with_') and (tmpl_context[k] in ['0', 'false', 'False']):
                tmpl_context[k] = False
        tmpl_context['ctx'] = self.get_object(ml)
        tmpl_context['cldf'] = self.dataset_mapping[ml.prefix]
        return tmpl_context

    def render_link(self, ml):
        ref_link = ml.all and \
            'cited_only' in ml.parsed_url_query and \
            ml.component(self.dataset_mapping[ml.prefix]) == 'Source'

        if self.cited:
            # We're in the second pass!
            if ref_link:
                ctx = self.get_tmpl_context(ml)
                ctx['ctx'] = [s for s in self.get_object(ml) if s.id in self.cited]
                ctx["with_anchor"] = True
                return self.render_template('Source', ctx, index=True)
            return ml

        if ref_link:  # So we need a second pass.
            self.with_partial_local_reflist = True
            return ml

        return self.render_template(
            ml.component(self.dataset_mapping[ml.prefix]) or ml.table_or_fname,
            self.get_tmpl_context(ml),
            index=ml.all)

    def render(self, simple_link_detection=True, markdown_kw=None):
        md = super().render(simple_link_detection=simple_link_detection, markdown_kw=markdown_kw)
        if self.with_partial_local_reflist:
            # Second pass!
            # 1. Determine which sources have been referenced:
            self.cited = set(
                MarkdownLink.from_match(ml).url.split('-', maxsplit=1)[-1]
                for ml in MarkdownLink.pattern.finditer(md)
                if MarkdownLink.from_match(ml).url.startswith('#source-'))

            # 2. Insert the pruned list of sources:
            md = super().render(
                simple_link_detection=simple_link_detection, markdown_kw=markdown_kw)
        return md



import webbrowser

from cldfviz.colormap import SHAPES

# For pacific-centered maps we chose 154°E as central longitude. This is particularly suitable,
# because the cut at 26°W does not cut through any macroareas.
# see https://en.wikipedia.org/wiki/154th_meridian_east and
# https://en.wikipedia.org/wiki/26th_meridian_west
PACIFIC_CENTERED = 154


class Map:
    __formats__ = []
    __marker_class__ = None

    def __init__(self, languages, args):
        self.languages = languages
        self.args = args

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        """write files"""
        raise NotImplementedError()

    @staticmethod
    def get_shape_and_color(colors_or_shapes):
        if len(colors_or_shapes) == 2:
            shapes, colors = [], []
            for c in colors_or_shapes:
                (shapes if c in SHAPES else colors).append(c)
            if shapes:
                if len(shapes) > 1:
                    raise ValueError('Only one shape can be specified for a marker')
                return shapes[0], colors[0]

    @staticmethod
    def add_options(parser, help_suffix):  # pragma: no cover
        pass

    def api_add_language(self, language, values, colormaps):  # pragma: no cover
        marker_spec = None
        if self.args.marker_factory:
            marker_spec = self.args.marker_factory(self, language, values, colormaps)
            if marker_spec is True:
                # All done!
                return
            assert isinstance(marker_spec, self.__marker_class__)
        self.add_language(language, values, colormaps, spec=marker_spec)

    def add_language(self, language, values, colormaps, spec=None):  # pragma: no cover
        raise NotImplementedError()

    def api_add_legend(self, parameters, colormaps):
        if self.args.marker_factory:
            return self.args.marker_factory.legend(self, parameters, colormaps)
        self.add_legend(parameters, colormaps)

    def add_legend(self, parameters, colormaps):  # pragma: no cover
        raise NotImplementedError()

    def open(self):  # pragma: no cover
        if self.args.format == 'html':
            webbrowser.open(self.args.output.resolve().as_uri(), new=1)



"""
Plot values for parameters in a CLDF StructureDataset on a map.

Usage examples:
- plot languages of a dataset:
  cldfbench cldfviz.map PATH/TO/DATASET --language-labels
- plot values of one parameter:
  cldfbench cldfviz.map PATH/TO/DATASET --parameters PID
- plot values of a column in the dataset's LanguageTable:
  cldfbench cldfviz.map PATH/TO/DATASET --language-property Macroarea

Colormaps: Colormaps can be specified by name - chosing from the ones available - or explicitly,
by providing a mapping from values (as found in the value column of ValueTable) to colors (see
below), serialized as JSON object, e.g. `--colormaps '{"x": "#a00", "y": "#0a0"}'`.

Colors: Colors can be specified as
- hex-triplets ("#a00", "AA0000")
- name (see https://www.w3.org/TR/css-color-4/#named-colors)
"""
import pathlib

from pycldf.cli_util import get_dataset, add_dataset
from clldutils.clilib import PathType, ParserError

from cldfviz.colormap import Colormap, COLORMAPS
from cldfviz.multiparameter import MultiParameter, CONTINUOUS, CATEGORICAL
from cldfviz.map import Map, MarkerFactory
from cldfviz.cli_util import (
    add_testable, add_listvalued, import_subclass, add_language_filter, get_language_filter)
from cldfviz.glottolog import Glottolog

FORMATS = {}
for cls in Map.__subclasses__():
    for fmt in cls.__formats__:
        FORMATS[fmt] = cls


def join_quoted(items):
    return ', '.join(['"{}"'.format(i) for i in items])


def register(parser):
    add_testable(parser)
    add_dataset(parser)
    add_language_filter(parser)
    Glottolog.add(parser)
    add_listvalued(
        parser,
        '--parameters',
        help="Comma-separated Parameter IDs, specifying the values to plot on the map. If not "
             "specified, all languages in the dataset will be plotted.",
    )
    add_listvalued(
        parser,
        '--datatypes',
        help="Explicit datatypes for parameters",
    )
    add_listvalued(
        parser,
        '--colormaps',
        help="Comma-separated names of colormaps to use for the respective parameter. Choose from "
             "{} for categorical and from {} for continuous parameters."
             "".format(join_quoted(COLORMAPS[CATEGORICAL]), join_quoted(COLORMAPS[CONTINUOUS])),
    )
    add_listvalued(
        parser,
        '--language-properties',
        help="Comma-separated language properties, i.e. columns in the dataset's LanguageTable "
             "to plot on the map.",
    )
    add_listvalued(
        parser,
        '--language-properties-colormaps',
        help="Comma-separated names of colormap to use for the respective language properties. "
             "See help for --colormaps for choices",
    )
    parser.add_argument(
        '--output',
        type=PathType(type='file', must_exist=False),
        help="Filesystem path to write the resulting map to. If no suffix is specified, it will "
             "be appended according to FORMAT; if given, suffix must match FORMAT.",
        default=pathlib.Path('map'))
    parser.add_argument(
        '--format',
        default='html',
        metavar='FORMAT',
        choices=list(FORMATS),
    )
    parser.add_argument(
        '--markersize',
        help="Size of map markers in pixels",
        type=int,
        default=10,
    )
    parser.add_argument(
        '--marker-factory',
        help="A python module providing a subclass of `cldfviz.map.MarkerFactory`.",
        default=None,
    )
    parser.add_argument(
        '--title',
        default=None,
        help="Title for the map plot",
    )
    parser.add_argument(
        '--pacific-centered',
        action='store_true',
        default=False,
        help="Center maps of the whole world at the pacific, thus not cutting large language "
             "families in half."
    )
    parser.add_argument(
        '--language-labels',
        action='store_true',
        default=False,
        help="Display language names on the map",
    )
    parser.add_argument(
        '--missing-value',
        default=None,
        help="A color used to indicate missing values. If not specified missing values will be "
             "omitted.",
    )
    parser.add_argument(
        '--no-legend',
        action='store_true',
        default=False,
        help="Don't add a legend to the map (e.g. because it would be too big).",
    )
    parser.add_argument(
        '--no-open',
        action='store_true',
        default=False,
        help="Don't open the created file.",
    )
    for cls in Map.__subclasses__():
        cls.add_options(
            parser, help_suffix='(Only for FORMATs {})'.format(join_quoted(cls.__formats__)))


def run(args):
    ds = get_dataset(args)
    if not args.output.suffix:
        args.output = args.output.parent / "{}.{}".format(args.output.name, args.format)
    else:
        assert args.output.suffix[1:] == args.format

    glottolog = Glottolog.from_args(args)
    data = MultiParameter(
        ds,
        args.parameters,
        datatypes=args.datatypes,
        glottolog=glottolog,
        include_missing=args.missing_value is not None,
        language_properties=args.language_properties,
        language_filter=get_language_filter(args),
    )
    if args.parameters and not args.colormaps:
        args.colormaps = [None] * len(args.parameters)
    if args.language_properties and not args.language_properties_colormaps:
        args.language_properties_colormaps = \
            [None] * len(args.language_properties)
    if '__language__' in data.parameters:
        assert len(data.parameters) == 1
        args.colormaps = [None]
    args.colormaps.extend(args.language_properties_colormaps)
    assert len(args.colormaps) == len(data.parameters), '{}'.format(data.parameters.keys())
    try:
        cms = {
            pid: Colormap(
                data.parameters[pid],
                name=cm,
                novalue=args.missing_value)
            for pid, cm in zip(data.parameters, args.colormaps)}
    except (ValueError, KeyError) as e:
        raise ParserError(str(e))

    with_shapes = sum(1 for cm in cms.values() if cm.with_shapes)
    if with_shapes:
        if with_shapes > 1:
            raise ParserError('Only one colormap can specify shapes.')
        if len(data.parameters) != 2:
            raise ParserError('Shapes can only be specified for one of two parameters.')

    if args.marker_factory:
        comps = args.marker_factory.split(',')
        cls = import_subclass(comps[0], MarkerFactory)
        args.marker_factory = cls(ds, args, *comps[1:])

    try:
        map = FORMATS[args.format](data.languages.values(), args)
    except ValueError as e:  # pragma: no cover
        raise ParserError(str(e))

    with map as fig:
        for lang, values in data.iter_languages():
            fig.api_add_language(lang, values, cms)

        if not args.no_legend:
            fig.api_add_legend(data.parameters, cms)

        args.log.info('Writing output to: {}'.format(args.output))
        args.log.info('For non-html maps this may take a while.')
        if args.test or args.no_open:
            return
        fig.open()  # pragma: no cover



import copy
import typing
import pathlib
import xml.etree.cElementTree as ElementTree

import toytree
import toyplot.svg
from pycldf.trees import Tree
from newick import RESERVED_PUNCTUATION, Node

__all__ = ['render']


def clean_node_label(s):
    for c in RESERVED_PUNCTUATION + ' ':
        s = s.replace(c, '_')
    return s


def render(tree: Tree,
           output: typing.Optional[pathlib.Path] = None,
           glottolog_mapping: typing.Optional[typing.Dict[str, typing.Tuple[str, str]]] = None,
           legend: typing.Optional[str] = None,
           width: typing.Optional[int] = 500,
           styles: typing.Optional[dict] = None,
           with_glottolog_links: bool = False,
           labels: typing.Optional[typing.Union[typing.Callable[[Node], str], dict]] = None,
           leafs: typing.Optional[typing.Union[typing.Callable[[Node], bool], list]] = None,
           ) -> typing.Union[pathlib.Path, str]:
    glottolog_mapping = glottolog_mapping or {}

    def rename(n):
        if n.name in glottolog_mapping:
            n.name = "{}--{}".format(n.name, glottolog_mapping[n.name][0])
        if not n.is_leaf:
            n.name = None

    def rename2(n):
        n.name = clean_node_label(labels(n) if callable(labels) else labels[n.name]) \
            if n.name else n.name

    nwk = tree.newick(strip_comments=True)
    if leafs:
        if callable(leafs):
            leafs = [n.name for n in nwk.walk() if n.name and leafs(n)]
        nwk.prune_by_names(leafs, inverse=True)
    if with_glottolog_links:
        nwk.visit(rename)
    if labels:
        nwk.visit(rename2)

    style = dict(
        width=width,
        height=sum(1 for n in nwk.walk() if n.is_leaf) * 15 + 50,
        node_hover=True,
        tip_labels_align=True,
        tip_labels_style={
            "fill": "#262626",
            "font-size": "11px",
            "-toyplot-anchor-shift": "5px",
            "line-height": "14px",
        },
        scalebar=bool(tree.tree_branch_length_unit) or bool(legend),
    )
    style.update(styles or {})
    canvas, axes, mark = toytree.tree(nwk.newick + ";", tree_format=1).draw(**style)
    if legend:
        axes.label.text = legend
    res = ElementTree.tostring(toyplot.svg.render(canvas, None)).decode('utf8')
    if with_glottolog_links:
        res = add_glottolog_links(res, glottolog_mapping)
    if output:
        output.write_text(res, encoding='utf8')
        return output
    return res


def add_glottolog_links(svg, gcodes, out=None):
    "Post-process the SVG to turn leaf names with Glottocodes into links"""
    ns = '{http://www.w3.org/2000/svg}'
    svg = ElementTree.fromstring(svg)
    for t in svg.findall('*.//{0}g[@class="toytree-TipLabels"]/{0}g/{0}text'.format(ns)):
        lid, _, gcode = t.text.strip().partition('--')
        if gcode:
            se = ElementTree.SubElement(t, '{0}text'.format(ns))
            gname = gcodes[lid][1]
            if gname:
                se.text = '{} - {} [{}]'.format(lid, gname, gcode)
            else:
                se.text = '{} - [{}]'.format(lid, gcode)
            se.attrib = copy.copy(t.attrib)
            se.attrib['fill'] = '#0000ff'
            t.tag = '{0}a'.format(ns)
            t.attrib = {
                'href': 'https://glottolog.org/resource/languoid/id/{}'.format(gcode),
                'title': 'The glottolog name',
            }
            t.text = None
    return ElementTree.tostring(svg).decode('utf8')



"""
Build an HTML file to display examples in a CLDF dataset.
"""
from pycldf.cli_util import add_dataset, get_dataset
from pycldf.sources import Sources

from cldfviz.cli_util import add_open, write_output, add_jinja_template
from cldfviz.template import render_jinja_template, TEMPLATE_DIR


def register(parser):
    add_dataset(parser)
    mod = __name__.split('.')[-1]
    add_jinja_template(parser, TEMPLATE_DIR / mod / '{}.html'.format(mod))
    add_open(parser)


def run(args):
    ds = get_dataset(args)
    res = render_jinja_template(
        args.template,
        ds=ds,
        examples=list(ds.objects("ExampleTable")),
        split_ref=Sources.parse)
    write_output(args, res)



"""
Utilities to bridge CLDF and pandas objects.
"""
import types
import typing
import collections

try:
    import pandas as pd
except ImportError:  # pragma: no cover
    pd = types.SimpleNamespace(DataFrame=object)

__all__ = ['df_from_dicts']


class DF:
    def __init__(self):
        self.df = None
        self.acc = collections.defaultdict(list)

    def __enter__(self):
        return self

    def add(self, item):
        if self.df is None:
            self.df = pd.DataFrame(columns=list(item))
        for k, v in item.items():
            self.acc[k].append(v)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for k, v in self.acc.items():
            self.df[k] = v


def df_from_dicts(dicts: typing.Iterable[dict]) -> pd.DataFrame:
    """
    Combines "rows" given as `dict`s into a pandas DataFrame.
    """
    with DF() as df:
        for d in dicts:
            df.add(d)
    return df.df
